@@ -2,32 +2,74 @@ use anyhow::Context as _;
 use std::collections::HashMap;
 use std::io;
 
+use terminal_size::terminal_size;
+
 use sequoia_net::pks;
 use sequoia_openpgp as openpgp;
-use crate::openpgp::types::SymmetricAlgorithm;
-use crate::openpgp::fmt::hex;
-use crate::openpgp::crypto::{self, SessionKey, Decryptor, Password};
-use crate::openpgp::{Fingerprint, Cert, KeyID, Result};
-use crate::openpgp::packet;
-use crate::openpgp::packet::prelude::*;
-use crate::openpgp::parse::{
+use openpgp::types::SymmetricAlgorithm;
+use openpgp::fmt::hex;
+use openpgp::crypto::{self, SessionKey, Decryptor, Password};
+use openpgp::{Fingerprint, Cert, KeyID, Result};
+use openpgp::packet;
+use openpgp::packet::prelude::*;
+use openpgp::parse::{
     Parse,
     PacketParser,
     PacketParserResult,
 };
-use crate::openpgp::parse::stream::{
+use openpgp::parse::stream::{
     VerificationHelper, DecryptionHelper, DecryptorBuilder, MessageStructure,
 };
 
 use crate::{
-    Config,
+    cli,
     commands::{
-        dump::PacketDumper,
+        toolbox::packet::dump::PacketDumper,
         VHelper,
     },
-    sq_cli::CliSessionKey,
+    common::password,
+    Config,
+    load_certs,
 };
 
+pub fn dispatch(config: Config, command: cli::decrypt::Command) -> Result<()> {
+    tracer!(TRACE, "decrypt::dispatch");
+
+    let mut input = command.input.open()?;
+    let mut output = command.output.create_safe(config.force)?;
+
+    let certs = load_certs(
+        command.sender_cert_file.iter().map(|s| s.as_ref()),
+    )?;
+    // Fancy default for --signatures.  If you change this,
+    // also change the description in the CLI definition.
+    let signatures = command.signatures.unwrap_or_else(|| {
+        if certs.is_empty() {
+            // No certs are given for verification, use 0 as
+            // threshold so we handle only-encrypted messages
+            // gracefully.
+            0
+        } else {
+            // At least one cert given, expect at least one
+            // valid signature.
+            1
+        }
+    });
+    // TODO: should this be load_keys?
+    let secrets =
+        load_certs(command.secret_key_file.iter().map(|s| s.as_ref()))?;
+    let private_key_store = command.private_key_store;
+    let session_keys = command.session_key;
+    decrypt(config, private_key_store.as_deref(),
+            &mut input, &mut output,
+            signatures, certs, secrets,
+            command.dump_session_key,
+            session_keys,
+            command.dump, command.hex)?;
+
+    Ok(())
+}
+
 trait PrivateKey {
     fn get_unlocked(&self) -> Option<Box<dyn Decryptor>>;
 
@@ -89,20 +131,20 @@ impl PrivateKey for RemotePrivateKey {
     }
 }
 
-struct Helper<'a> {
-    vhelper: VHelper<'a>,
+struct Helper<'a, 'certdb> {
+    vhelper: VHelper<'a, 'certdb>,
     secret_keys: HashMap<KeyID, Box<dyn PrivateKey>>,
     key_identities: HashMap<KeyID, Fingerprint>,
     key_hints: HashMap<KeyID, String>,
-    session_keys: Vec<CliSessionKey>,
+    session_keys: Vec<cli::types::SessionKey>,
     dump_session_key: bool,
     dumper: Option<PacketDumper>,
 }
 
-impl<'a> Helper<'a> {
-    fn new(config: &Config<'a>, private_key_store: Option<&str>,
+impl<'a, 'certdb> Helper<'a, 'certdb> {
+    fn new(config: &'a Config<'certdb>, private_key_store: Option<&str>,
            signatures: usize, certs: Vec<Cert>, secrets: Vec<Cert>,
-           session_keys: Vec<CliSessionKey>,
+           session_keys: Vec<cli::types::SessionKey>,
            dump_session_key: bool, dump: bool)
            -> Self
     {
@@ -147,7 +189,8 @@ impl<'a> Helper<'a> {
             session_keys,
             dump_session_key,
             dumper: if dump {
-                let width = term_size::dimensions_stdout().map(|(w, _)| w)
+                let width = terminal_size()
+                    .map(|(width, _height)| width.0.into())
                     .unwrap_or(80);
                 Some(PacketDumper::new(width, false))
             } else {
@@ -173,7 +216,7 @@ impl<'a> Helper<'a> {
         {
             Some(sk) => {
                 if self.dump_session_key {
-                    eprintln!("Session key: {}", hex::encode(&sk));
+                    wprintln!("Session key: {}", hex::encode(&sk));
                 }
                 Some(self.key_identities.get(&keyid).cloned())
             },
@@ -182,13 +225,13 @@ impl<'a> Helper<'a> {
     }
 }
 
-impl<'a> VerificationHelper for Helper<'a> {
+impl<'a, 'certdb> VerificationHelper for Helper<'a, 'certdb> {
     fn inspect(&mut self, pp: &PacketParser) -> Result<()> {
         if let Some(dumper) = self.dumper.as_mut() {
             dumper.packet(&mut io::stderr(),
                           pp.recursion_depth() as usize,
                           pp.header().clone(), pp.packet.clone(),
-                          pp.map().cloned(), None)?;
+                          pp.map().cloned(), Vec::new())?;
         }
         Ok(())
     }
@@ -201,8 +244,7 @@ impl<'a> VerificationHelper for Helper<'a> {
     }
 }
 
-impl<'a> DecryptionHelper for Helper<'a> {
-    #[allow(clippy::if_let_some_result)]
+impl<'a, 'certdb> DecryptionHelper for Helper<'a, 'certdb> {
     fn decrypt<D>(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
                   sym_algo: Option<SymmetricAlgorithm>,
                   mut decrypt: D) -> openpgp::Result<Option<Fingerprint>>
@@ -220,7 +262,7 @@ impl<'a> DecryptionHelper for Helper<'a> {
                     .any(|sa| decrypt(sa, &sk.session_key))
             };
             if decrypted {
-                eprintln!("Encrypted with Session Key {}", sk.display_sensitive());
+                wprintln!("Encrypted with Session Key {}", sk.display_sensitive());
                 return Ok(None);
             }
         }
@@ -254,14 +296,14 @@ impl<'a> DecryptionHelper for Helper<'a> {
                         break keypair;
                     }
 
-                    let p = rpassword::read_password_from_tty(Some(
-                        &format!(
-                            "Enter password to decrypt key {}: ",
-                            self.key_hints.get(keyid).unwrap())))?.into();
+                    let p = password::prompt_to_unlock(&format!(
+                        "key {}",
+                        self.key_hints.get(keyid).unwrap()
+                    ))?;
 
                     match key.unlock(&p) {
                         Ok(decryptor) => break decryptor,
-                        Err(error) => eprintln!("Could not unlock key: {:?}", error),
+                        Err(error) => wprintln!("Could not unlock key: {:?}", error),
                     }
                 };
 
@@ -308,15 +350,15 @@ impl<'a> DecryptionHelper for Helper<'a> {
                         break keypair;
                     }
 
-                    let p = rpassword::read_password_from_tty(Some(
-                        &format!(
-                            "Enter password to decrypt key {}: ",
-                            self.key_hints.get(&keyid).unwrap())))?.into();
+                    let p = password::prompt_to_unlock(&format!(
+                        "key {}",
+                        self.key_hints.get(&keyid).unwrap()
+                    ))?;
 
                     if let Ok(decryptor) = key.unlock(&p) {
                         break decryptor;
                     } else {
-                        eprintln!("Bad password.");
+                        wprintln!("Bad password.");
                     }
                 };
 
@@ -336,22 +378,20 @@ impl<'a> DecryptionHelper for Helper<'a> {
 
         // Finally, try to decrypt using the SKESKs.
         loop {
-            let password =
-                rpassword::read_password_from_tty(Some(
-                    "Enter password to decrypt message: "))?.into();
+            let password = password::prompt_to_unlock("message")?;
 
             for skesk in skesks {
                 if let Some(sk) = skesk.decrypt(&password).ok()
                     .and_then(|(algo, sk)| { if decrypt(algo, &sk) { Some(sk) } else { None }})
                 {
                     if self.dump_session_key {
-                        eprintln!("Session key: {}", hex::encode(&sk));
+                        wprintln!("Session key: {}", hex::encode(&sk));
                     }
                     return Ok(None);
                 }
             }
 
-            eprintln!("Bad password.");
+            wprintln!("Incorrect password.");
         }
     }
 }
@@ -364,7 +404,7 @@ pub fn decrypt(config: Config,
                output: &mut dyn io::Write,
                signatures: usize, certs: Vec<Cert>, secrets: Vec<Cert>,
                dump_session_key: bool,
-               sk: Vec<CliSessionKey>,
+               sk: Vec<cli::types::SessionKey>,
                dump: bool, hex: bool)
                -> Result<()> {
     let helper = Helper::new(&config, private_key_store, signatures, certs,
@@ -388,7 +428,7 @@ pub fn decrypt_unwrap(config: Config,
                       input: &mut (dyn io::Read + Sync + Send),
                       output: &mut dyn io::Write,
                       secrets: Vec<Cert>,
-                      session_keys: Vec<CliSessionKey>,
+                      session_keys: Vec<cli::types::SessionKey>,
                       dump_session_key: bool)
                       -> Result<()>
 {
@@ -425,6 +465,7 @@ pub fn decrypt_unwrap(config: Config,
                 io::copy(&mut pp, output)?;
                 return Ok(());
             },
+            #[allow(deprecated)]
             Packet::MDC(ref mdc) => if ! mdc.valid() {
                 return Err(openpgp::Error::ManipulatedMessage.into());
             },
