@@ -1,31 +1,48 @@
 use std::convert::TryFrom;
 use std::io::{self, Read};
+use std::time::{Duration, SystemTime};
+
+use anyhow::Context;
 
 use sequoia_openpgp as openpgp;
-use crate::openpgp::{KeyHandle, Packet, Result};
-use crate::openpgp::cert::prelude::*;
+use openpgp::{KeyHandle, Packet, Result};
+use openpgp::cert::prelude::*;
 use openpgp::packet::{
     Signature,
     key::PublicParts,
 };
-use crate::openpgp::parse::{Parse, PacketParserResult};
-use crate::openpgp::policy::{Policy, HashAlgoSecurity};
-use crate::openpgp::packet::key::SecretKeyMaterial;
+use openpgp::parse::{Parse, PacketParserResult};
+use openpgp::policy::{Policy, HashAlgoSecurity};
+use openpgp::packet::key::SecretKeyMaterial;
+use openpgp::types::SignatureType;
 
-use super::dump::Convert;
+use sequoia_cert_store as cert_store;
+use cert_store::Store;
 
+use crate::Convert;
+
+use crate::Config;
 use crate::SECONDS_IN_YEAR;
 use crate::SECONDS_IN_DAY;
 
-use crate::sq_cli::InspectCommand;
+use crate::cli::inspect;
+use crate::cli::types::FileOrStdout;
+
+pub fn dispatch(config: Config, c: inspect::Command)
+    -> Result<()>
+{
+    // sq inspect does not have --output, but commands::inspect does.
+    // Work around this mismatch by always creating a stdout output.
+    let output_type = FileOrStdout::default();
+    let output = &mut output_type.create_unsafe(config.force)?;
+
+    let policy = &config.policy;
+    let time = Some(config.time);
 
-pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Write)
-               -> Result<()> {
     let print_certifications = c.certifications;
 
-    let input = c.input.as_deref();
-    let input_name = input.unwrap_or("-");
-    write!(output, "{}: ", input_name)?;
+    let input = c.input;
+    write!(output, "{}: ", input)?;
 
     let mut type_called = false;  // Did we print the type yet?
     let mut encrypted = false;    // Is it an encrypted message?
@@ -35,8 +52,34 @@ pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Writ
     let mut sigs = Vec::new();    // Accumulator for signatures.
     let mut literal_prefix = Vec::new();
 
-    let mut ppr =
-        openpgp::parse::PacketParser::from_reader(crate::open_or_stdin(input)?)?;
+    let mut bytes: Vec<u8> = Vec::new();
+    let mut ppr = if c.cert.is_empty() {
+        if let Some(path) = input.inner() {
+            if ! path.exists() &&
+                format!("{}", input).parse::<KeyHandle>().is_ok() {
+                wprintln!("The file {} does not exist, \
+                           did you mean \"sq inspect --cert {}\"?",
+                          input, input);
+            }
+        }
+
+        openpgp::parse::PacketParser::from_reader(input.open()?)?
+    } else {
+        let cert_store = config.cert_store_or_else()?;
+        for cert in c.cert.into_iter() {
+            let certs = cert_store.lookup_by_cert_or_subkey(&cert)
+                .with_context(|| format!("Looking up {}", cert))?;
+
+            // Include non-exportable signatures, etc.
+            for cert in certs.into_iter() {
+                let b = cert.to_vec().context("Serializing certificate")?;
+                bytes.extend(b);
+            }
+        }
+
+        openpgp::parse::PacketParser::from_bytes(&bytes)?
+    };
+
     while let PacketParserResult::Some(mut pp) = ppr {
         match pp.packet {
             Packet::PublicKey(_) | Packet::SecretKey(_) => {
@@ -51,8 +94,13 @@ pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Writ
                     let pp = openpgp::PacketPile::from(
                         std::mem::take(&mut packets));
                     let cert = openpgp::Cert::try_from(pp)?;
-                    inspect_cert(policy, output, &cert,
-                                 print_certifications)?;
+                    inspect_cert(
+                        policy,
+                        time,
+                        output,
+                        &cert,
+                        print_certifications,
+                    )?;
                 }
             },
             Packet::Literal(_) => {
@@ -110,13 +158,23 @@ pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Writ
         } else if is_cert.is_ok() || is_keyring.is_ok() {
             let pp = openpgp::PacketPile::from(packets);
             let cert = openpgp::Cert::try_from(pp)?;
-            inspect_cert(policy, output, &cert,
-                         print_certifications)?;
+            inspect_cert(policy, time, output, &cert, print_certifications)?;
         } else if packets.is_empty() && ! sigs.is_empty() {
-            writeln!(output, "Detached signature{}.",
-                     if sigs.len() > 1 { "s" } else { "" })?;
-            writeln!(output)?;
-            inspect_signatures(output, &sigs)?;
+            if sigs.iter().all(is_revocation_sig) {
+                writeln!(output, "Revocation Certificate{}.",
+                         if sigs.len() > 1 { "s" } else { "" })?;
+                writeln!(output)?;
+                for sig in sigs {
+                    inspect_bare_revocation(output, &sig)?;
+                }
+                writeln!(output, "           Note: \
+                                  Signatures have NOT been verified!")?;
+            } else {
+                writeln!(output, "Detached signature{}.",
+                         if sigs.len() > 1 { "s" } else { "" })?;
+                writeln!(output)?;
+                inspect_signatures(output, &sigs)?;
+            }
         } else if packets.is_empty() {
             writeln!(output, "No OpenPGP data.")?;
         } else {
@@ -125,7 +183,7 @@ pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Writ
             writeln!(output, "  Cert: {}", is_cert.unwrap_err())?;
             writeln!(output, "  Keyring: {}", is_keyring.unwrap_err())?;
             writeln!(output)?;
-            writeln!(output, "Hint: Try 'sq packet dump {}'", input_name)?;
+            writeln!(output, "Hint: Try 'sq packet dump {}'", input)?;
         }
     } else {
         unreachable!()
@@ -134,26 +192,72 @@ pub fn inspect(c: InspectCommand, policy: &dyn Policy, output: &mut dyn io::Writ
     Ok(())
 }
 
-fn inspect_cert(policy: &dyn Policy,
-                output: &mut dyn io::Write, cert: &openpgp::Cert,
-                print_certifications: bool) -> Result<()> {
+/// Returns true iff all signatures in the cert are revocation
+/// signatures.
+fn is_revocation_sig(s: &Signature) -> bool {
+    [
+        SignatureType::KeyRevocation,
+        SignatureType::SubkeyRevocation,
+        SignatureType::CertificationRevocation,
+    ].contains(&s.typ())
+}
+
+/// Returns true iff all signatures in the cert are revocation
+/// signatures.
+fn is_revocation_cert(c: &Cert) -> bool {
+    c.primary_key().signatures().all(|s| s.typ() == SignatureType::KeyRevocation)
+        && c.keys().subkeys().all(|skb| skb.signatures().all(
+            |s| s.typ() == SignatureType::SubkeyRevocation))
+        && c.userids().all(|uidb| uidb.signatures().all(
+            |s| s.typ() == SignatureType::CertificationRevocation))
+}
+
+fn inspect_cert(
+    policy: &dyn Policy,
+    time: Option<SystemTime>,
+    output: &mut dyn io::Write,
+    cert: &openpgp::Cert,
+    print_certifications: bool,
+) -> Result<()> {
     if cert.is_tsk() {
         writeln!(output, "Transferable Secret Key.")?;
+    } else if is_revocation_cert(&cert) {
+        writeln!(output, "Revocation Certificate.")?;
     } else {
         writeln!(output, "OpenPGP Certificate.")?;
     }
     writeln!(output)?;
     writeln!(output, "    Fingerprint: {}", cert.fingerprint())?;
     inspect_revocation(output, "", cert.revocation_status(policy, None))?;
-    inspect_key(policy, output, "", cert.keys().next().unwrap(),
-                print_certifications)?;
+    inspect_key(
+        policy,
+        time,
+        output,
+        "",
+        cert.keys().next().unwrap(),
+        print_certifications,
+    )?;
     writeln!(output)?;
 
-    for vka in cert.keys().subkeys().with_policy(policy, None) {
-        writeln!(output, "         Subkey: {}", vka.key().fingerprint())?;
-        inspect_revocation(output, "", vka.revocation_status())?;
-        inspect_key(policy, output, "", vka.into_key_amalgamation().into(),
-                    print_certifications)?;
+    for skb in cert.keys().subkeys() {
+        writeln!(output, "         Subkey: {}", skb.key().fingerprint())?;
+        inspect_revocation(output, "", skb.revocation_status(policy, None))?;
+        match skb.binding_signature(policy, None) {
+            Ok(sig) => {
+                if let Err(e) = sig.signature_alive(None, Duration::new(0, 0)) {
+                    print_error_chain(output, &e)?;
+                }
+            }
+            Err(e) => print_error_chain(output, &e)?,
+        }
+        inspect_key(
+            policy,
+            time,
+            output,
+            "",
+            skb.into(),
+            print_certifications,
+        )?;
         writeln!(output)?;
     }
 
@@ -170,10 +274,10 @@ fn inspect_cert(policy: &dyn Policy,
         writeln!(output, "         UserID: {}", uidb.userid())?;
         inspect_revocation(output, "", uidb.revocation_status(policy, None))?;
         match uidb.binding_signature(policy, None) {
-            Ok(sig) => if let Err(e) =
-                sig.signature_alive(None, std::time::Duration::new(0, 0))
-            {
-                print_error_chain(output, &e)?;
+            Ok(sig) => {
+                if let Err(e) = sig.signature_alive(None, Duration::new(0, 0)) {
+                    print_error_chain(output, &e)?;
+                }
             }
             Err(e) => print_error_chain(output, &e)?,
         }
@@ -188,10 +292,10 @@ fn inspect_cert(policy: &dyn Policy,
                  uab.user_attribute())?;
         inspect_revocation(output, "", uab.revocation_status(policy, None))?;
         match uab.binding_signature(policy, None) {
-            Ok(sig) => if let Err(e) =
-                sig.signature_alive(None, std::time::Duration::new(0, 0))
-            {
-                print_error_chain(output, &e)?;
+            Ok(sig) => {
+                if let Err(e) = sig.signature_alive(None, Duration::new(0, 0)) {
+                    print_error_chain(output, &e)?;
+                }
             }
             Err(e) => print_error_chain(output, &e)?,
         }
@@ -204,10 +308,10 @@ fn inspect_cert(policy: &dyn Policy,
     for ub in cert.unknowns() {
         writeln!(output, "         Unknown component: {:?}", ub.unknown())?;
         match ub.binding_signature(policy, None) {
-            Ok(sig) => if let Err(e) =
-                sig.signature_alive(None, std::time::Duration::new(0, 0))
-            {
-                print_error_chain(output, &e)?;
+            Ok(sig) => {
+                if let Err(e) = sig.signature_alive(None, Duration::new(0, 0)) {
+                    print_error_chain(output, &e)?;
+                }
             }
             Err(e) => print_error_chain(output, &e)?,
         }
@@ -224,16 +328,17 @@ fn inspect_cert(policy: &dyn Policy,
     Ok(())
 }
 
-fn inspect_key(policy: &dyn Policy,
-               output: &mut dyn io::Write,
-               indent: &str,
-               ka: ErasedKeyAmalgamation<PublicParts>,
-               print_certifications: bool)
-        -> Result<()>
-{
+fn inspect_key(
+    policy: &dyn Policy,
+    time: Option<SystemTime>,
+    output: &mut dyn io::Write,
+    indent: &str,
+    ka: ErasedKeyAmalgamation<PublicParts>,
+    print_certifications: bool,
+) -> Result<()> {
     let key = ka.key();
     let bundle = ka.bundle();
-    let vka = match ka.with_policy(policy, None) {
+    let vka = match ka.with_policy(policy, time) {
         Ok(vka) => {
             if let Err(e) = vka.alive() {
                 writeln!(output, "{}                 Invalid: {}", indent, e)?;
@@ -281,48 +386,50 @@ fn inspect_key(policy: &dyn Policy,
     Ok(())
 }
 
+/// Prints the revocation reasons.
+fn print_reasons(output: &mut dyn io::Write, indent: &str,
+                 third_party: bool, sigs: &[&Signature])
+                 -> Result<()> {
+    for sig in sigs {
+        if let Some((r, msg)) = sig.reason_for_revocation() {
+            writeln!(output, "{}                  - {}", indent, r)?;
+            if third_party {
+                writeln!(output, "{}                    Issued by {}",
+                         indent,
+                         if let Some(issuer)
+                         = sig.get_issuers().into_iter().next()
+                         {
+                             issuer.to_string()
+                         } else {
+                             "an unknown certificate".into()
+                         })?;
+            }
+            writeln!(output, "{}                    Message: {:?}",
+                     indent, String::from_utf8_lossy(msg))?;
+        } else {
+            writeln!(output, "{}                  - No reason specified",
+                     indent)?;
+            if third_party {
+                writeln!(output, "{}                    Issued by {}",
+                         indent,
+                         if let Some(issuer)
+                         = &sig.get_issuers().into_iter().next()
+                         {
+                             issuer.to_string()
+                         } else {
+                             "an unknown certificate".into()
+                         })?;
+            }
+        }
+    }
+    Ok(())
+}
+
 fn inspect_revocation(output: &mut dyn io::Write,
                       indent: &str,
                       revoked: openpgp::types::RevocationStatus)
                       -> Result<()> {
-    use crate::openpgp::types::RevocationStatus::*;
-    fn print_reasons(output: &mut dyn io::Write, indent: &str,
-                     third_party: bool, sigs: &[&Signature])
-                     -> Result<()> {
-        for sig in sigs {
-            if let Some((r, msg)) = sig.reason_for_revocation() {
-                writeln!(output, "{}                  - {}", indent, r)?;
-                if third_party {
-                    writeln!(output, "{}                    Issued by {}",
-                             indent,
-                             if let Some(issuer)
-                                 = sig.get_issuers().into_iter().next()
-                             {
-                                 issuer.to_string()
-                             } else {
-                                 "an unknown certificate".into()
-                             })?;
-                }
-                writeln!(output, "{}                    Message: {:?}",
-                         indent, String::from_utf8_lossy(msg))?;
-            } else {
-                writeln!(output, "{}                  - No reason specified",
-                         indent)?;
-                if third_party {
-                    writeln!(output, "{}                    Issued by {}",
-                             indent,
-                             if let Some(issuer)
-                                 = &sig.get_issuers().into_iter().next()
-                             {
-                                 issuer.to_string()
-                             } else {
-                                 "an unknown certificate".into()
-                             })?;
-                }
-            }
-        }
-        Ok(())
-    }
+    use openpgp::types::RevocationStatus::*;
     match revoked {
         Revoked(sigs) => {
             writeln!(output, "{}                 Revoked:", indent)?;
@@ -338,6 +445,16 @@ fn inspect_revocation(output: &mut dyn io::Write,
     Ok(())
 }
 
+fn inspect_bare_revocation(output: &mut dyn io::Write, sig: &Signature)
+                           -> Result<()> {
+    let indent = "";
+    inspect_issuers(output, &sig)?;
+    writeln!(output, "{}                 Possible revocation:", indent)?;
+    print_reasons(output, indent, false, &[sig])?;
+    writeln!(output)?;
+    Ok(())
+}
+
 fn inspect_key_flags(flags: openpgp::types::KeyFlags) -> Option<String> {
     let mut capabilities = Vec::new();
     if flags.for_certification() {
@@ -371,7 +488,7 @@ fn inspect_key_flags(flags: openpgp::types::KeyFlags) -> Option<String> {
 
 fn inspect_signatures(output: &mut dyn io::Write,
                       sigs: &[openpgp::packet::Signature]) -> Result<()> {
-    use crate::openpgp::types::SignatureType::*;
+    use openpgp::types::SignatureType::*;
     for sig in sigs {
         match sig.typ() {
             Binary | Text => (),
@@ -379,21 +496,7 @@ fn inspect_signatures(output: &mut dyn io::Write,
                 writeln!(output, "           Kind: {}", signature_type)?,
         }
 
-        let mut fps: Vec<_> = sig.issuer_fingerprints().collect();
-        fps.sort();
-        fps.dedup();
-        let fps: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
-        for fp in fps.iter() {
-            writeln!(output, " Alleged signer: {}", fp)?;
-        }
-        let mut keyids: Vec<_> = sig.issuers().collect();
-        keyids.sort();
-        keyids.dedup();
-        for keyid in keyids {
-            if ! fps.iter().any(|fp| fp.aliases(&keyid.into())) {
-                writeln!(output, " Alleged signer: {}", keyid)?;
-            }
-        }
+        inspect_issuers(output, &sig)?;
     }
     if ! sigs.is_empty() {
         writeln!(output, "           Note: \
@@ -403,6 +506,28 @@ fn inspect_signatures(output: &mut dyn io::Write,
     Ok(())
 }
 
+fn inspect_issuers(output: &mut dyn io::Write,
+                   sig: &Signature) -> Result<()> {
+    let mut fps: Vec<_> = sig.issuer_fingerprints().collect();
+    fps.sort();
+    fps.dedup();
+    let fps: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
+    for fp in fps.iter() {
+        writeln!(output, " Alleged signer: {}", fp)?;
+    }
+
+    let mut keyids: Vec<_> = sig.issuers().collect();
+    keyids.sort();
+    keyids.dedup();
+    for keyid in keyids {
+        if ! fps.iter().any(|fp| fp.aliases(&keyid.into())) {
+            writeln!(output, " Alleged signer: {}", keyid)?;
+        }
+    }
+
+    Ok(())
+}
+
 fn inspect_certifications<'a, A>(output: &mut dyn io::Write,
                                  policy: &dyn Policy,
                                  certs: A,
@@ -503,7 +628,7 @@ fn inspect_certifications<'a, A>(output: &mut dyn io::Write,
                 sig, HashAlgoSecurity::CollisionResistance)
             {
                 writeln!(output,
-                         "{}Certificate is not valid according to \
+                         "{}Certification is not valid according to \
                           the current policy:\n\
                           {}  {}",
                          indent, indent, err)?;
