@@ -6,65 +6,141 @@ use std::time::SystemTime;
 use tempfile::NamedTempFile;
 
 use sequoia_openpgp as openpgp;
-use crate::openpgp::armor;
-use crate::openpgp::{Packet, Result};
-use crate::openpgp::packet::prelude::*;
-use crate::openpgp::packet::signature::subpacket::NotationData;
-use crate::openpgp::parse::{
+use openpgp::armor;
+use openpgp::{Packet, Result};
+use openpgp::packet::prelude::*;
+use openpgp::packet::signature::subpacket::NotationData;
+use openpgp::parse::{
     Parse,
     PacketParserResult,
 };
-use crate::openpgp::serialize::Serialize;
-use crate::openpgp::serialize::stream::{
+use openpgp::serialize::Serialize;
+use openpgp::serialize::stream::{
     Message, Armorer, Signer, LiteralWriter,
 };
-use crate::openpgp::types::SignatureType;
-use crate::{
-    Config,
-};
+use openpgp::types::SignatureType;
 
-pub struct SignOpts<'a> {
-    pub config: Config<'a>,
-    pub private_key_store: Option<&'a str>,
-    pub input: &'a mut (dyn io::Read + Sync + Send),
-    pub output_path: Option<&'a str>,
-    pub secrets: Vec<openpgp::Cert>,
-    pub detached: bool,
-    pub binary: bool,
-    pub append: bool,
-    pub notarize: bool,
-    pub time: Option<SystemTime>,
-    pub notations: &'a [(bool, NotationData)]
+use crate::Config;
+use crate::load_certs;
+use crate::parse_notations;
+
+use crate::cli;
+use crate::cli::types::FileOrStdin;
+use crate::cli::types::FileOrStdout;
+
+mod merge_signatures;
+use merge_signatures::merge_signatures;
+
+pub fn dispatch(config: Config, command: cli::sign::Command) -> Result<()> {
+    tracer!(TRACE, "sign::dispatch");
+
+    let mut input = command.input.open()?;
+    let output = &command.output;
+    let detached = command.detached;
+    let binary = command.binary;
+    let append = command.append;
+    let notarize = command.notarize;
+    if notarize {
+        return Err(anyhow::anyhow!("Notarizing messages is not supported."));
+    }
+
+    let private_key_store = command.private_key_store.as_deref();
+    let secrets =
+        load_certs(command.secret_key_file.iter().map(|s| s.as_ref()))?;
+    let time = Some(config.time);
+
+    let notations = parse_notations(command.notation)?;
+
+    if let Some(merge) = command.merge {
+        let output = output.create_pgp_safe(
+            config.force,
+            binary,
+            armor::Kind::Message,
+        )?;
+        let data: FileOrStdin = merge.into();
+        let mut input2 = data.open()?;
+        merge_signatures(&mut input, &mut input2, output)?;
+    } else if command.clearsign {
+        let output = output.create_safe(config.force)?;
+        clearsign(config, private_key_store, input, output, secrets,
+                  time, &notations)?;
+    } else {
+        sign(config,
+             private_key_store,
+             &mut input,
+             output,
+             secrets,
+             detached,
+             binary,
+             append,
+             notarize,
+             time,
+             &notations)?;
+    }
+
+    Ok(())
 }
 
-pub fn sign(opts: SignOpts) -> Result<()> {
-    match (opts.detached, opts.append|opts.notarize) {
+pub fn sign<'a, 'certdb>(
+    config: Config<'certdb>,
+    private_key_store: Option<&'a str>,
+    input: &'a mut (dyn io::Read + Sync + Send),
+    output: &'a FileOrStdout,
+    secrets: Vec<openpgp::Cert>,
+    detached: bool,
+    binary: bool,
+    append: bool,
+    notarize: bool,
+    time: Option<SystemTime>,
+    notations: &'a [(bool, NotationData)])
+    -> Result<()>
+{
+    match (detached, append|notarize) {
         (_, false) | (true, true) =>
-            sign_data(opts),
+            sign_data(config,
+                      private_key_store,
+                      input,
+                      output,
+                      secrets,
+                      detached,
+                      binary,
+                      append,
+                      time,
+                      notations),
         (false, true) =>
-            sign_message(opts),
+            sign_message(config,
+                         private_key_store,
+                         input,
+                         output,
+                         secrets,
+                         binary,
+                         notarize,
+                         time,
+                         notations),
     }
 }
 
-fn sign_data(opts: SignOpts) -> Result<()> {
-    let SignOpts {
-        config,
-        private_key_store,
-        input,
-        output_path,
-        secrets,
-        detached,
-        binary,
-        append,
-        time,
-        notations, ..} = opts;
+fn sign_data<'a, 'certdb>(
+    config: Config<'certdb>,
+    private_key_store: Option<&'a str>,
+    input: &'a mut (dyn io::Read + Sync + Send),
+    output_path: &'a FileOrStdout,
+    secrets: Vec<openpgp::Cert>,
+    detached: bool,
+    binary: bool,
+    append: bool,
+    time: Option<SystemTime>,
+    notations: &'a [(bool, NotationData)])
+    -> Result<()>
+{
     let (mut output, prepend_sigs, tmp_path):
     (Box<dyn io::Write + Sync + Send>, Vec<Signature>, Option<PathBuf>) =
-        if detached && append && output_path.is_some() {
+        if detached && append && output_path.path().is_some() {
+            let output_path = output_path.path().unwrap();
             // First, read the existing signatures.
             let mut sigs = Vec::new();
             let mut ppr =
-                openpgp::parse::PacketParser::from_file(output_path.unwrap())?;
+                openpgp::parse::PacketParser::from_file(output_path)?;
 
             while let PacketParserResult::Some(pp) = ppr {
                 let (packet, ppr_tmp) = pp.recurse()?;
@@ -83,12 +159,12 @@ fn sign_data(opts: SignOpts) -> Result<()> {
             // successful with adding our signature(s), we rename the
             // file replacing the old one.
             let tmp_file = NamedTempFile::new_in(
-                PathBuf::from(output_path.unwrap()).parent()
+                PathBuf::from(output_path).parent()
                     .unwrap_or(&PathBuf::from(".")))?;
             let tmp_path = tmp_file.path().into();
             (Box::new(tmp_file), sigs, Some(tmp_path))
         } else {
-            (config.create_or_stdout_safe(output_path)?, Vec::new(), None)
+            (output_path.create_safe(config.force)?, Vec::new(), None)
         };
 
     let mut keypairs = super::get_signing_keys(
@@ -127,12 +203,12 @@ fn sign_data(opts: SignOpts) -> Result<()> {
     }
 
     let mut signer = Signer::with_template(
-        message, keypairs.pop().unwrap(), builder);
+        message, keypairs.pop().unwrap().0, builder);
     if let Some(time) = time {
         signer = signer.creation_time(time);
     }
     for s in keypairs {
-        signer = signer.add_signer(s);
+        signer = signer.add_signer(s.0);
     }
     if detached {
         signer = signer.detached();
@@ -158,33 +234,54 @@ fn sign_data(opts: SignOpts) -> Result<()> {
 
     if let Some(path) = tmp_path {
         // Atomically replace the old file.
-        fs::rename(path,
-                   output_path.expect("must be Some if tmp_path is Some"))?;
+        fs::rename(
+            path,
+            output_path.path().expect("must be Some if tmp_path is Some"),
+        )?;
     }
     Ok(())
 }
 
-fn sign_message(opts: SignOpts) -> Result<()> {
-    let mut output =
-        opts.config.create_or_stdout_pgp(opts.output_path,
-                                    opts.binary,
-                                    armor::Kind::Message)?;
-    sign_message_(opts, &mut output)?;
+fn sign_message<'a, 'certdb>(
+    config: Config<'certdb>,
+    private_key_store: Option<&'a str>,
+    input: &'a mut (dyn io::Read + Sync + Send),
+    output: &'a FileOrStdout,
+    secrets: Vec<openpgp::Cert>,
+    binary: bool,
+    notarize: bool,
+    time: Option<SystemTime>,
+    notations: &'a [(bool, NotationData)])
+    -> Result<()>
+{
+    let mut output = output.create_pgp_safe(
+        config.force,
+        binary,
+        armor::Kind::Message,
+    )?;
+    sign_message_(config,
+                  private_key_store,
+                  input,
+                  secrets,
+                  notarize,
+                  time,
+                  notations,
+                  &mut output)?;
     output.finalize()?;
     Ok(())
 }
 
-fn sign_message_(opts: SignOpts, output: &mut (dyn io::Write + Sync + Send)) -> Result<()>
+fn sign_message_<'a, 'certdb>(
+    config: Config<'certdb>,
+    private_key_store: Option<&'a str>,
+    input: &'a mut (dyn io::Read + Sync + Send),
+    secrets: Vec<openpgp::Cert>,
+    notarize: bool,
+    time: Option<SystemTime>,
+    notations: &'a [(bool, NotationData)],
+    output: &mut (dyn io::Write + Sync + Send))
+    -> Result<()>
 {
-    let SignOpts {
-        config,
-        private_key_store,
-        input,
-        secrets,
-        notarize,
-        time,
-        notations, ..
-    } = opts;
     let mut keypairs = super::get_signing_keys(
         &secrets, &config.policy, private_key_store, time, None)?;
     if keypairs.is_empty() {
@@ -265,12 +362,12 @@ fn sign_message_(opts: SignOpts, output: &mut (dyn io::Write + Sync + Send)) ->
                 }
 
                 let mut signer = Signer::with_template(
-                    sink, keypairs.pop().unwrap(), builder);
+                    sink, keypairs.pop().unwrap().0, builder);
                 if let Some(time) = time {
                     signer = signer.creation_time(time);
                 }
                 for s in keypairs.drain(..) {
-                    signer = signer.add_signer(s);
+                    signer = signer.add_signer(s.0);
                 }
                 sink = signer.build().context("Failed to create signer")?;
                 state = State::Signing { signature_count: 0, };
@@ -410,13 +507,13 @@ pub fn clearsign(config: Config,
 
     let message = Message::new(&mut output);
     let mut signer = Signer::with_template(
-        message, keypairs.pop().unwrap(), builder)
+        message, keypairs.pop().unwrap().0, builder)
         .cleartext();
     if let Some(time) = time {
         signer = signer.creation_time(time);
     }
     for s in keypairs {
-        signer = signer.add_signer(s);
+        signer = signer.add_signer(s.0);
     }
     let mut message = signer.build().context("Failed to create signer")?;
 
