@@ -12,9 +12,10 @@ use openpgp::cert::prelude::*;
 use openpgp::KeyHandle;
 use openpgp::packet::signature::subpacket::NotationData;
 use openpgp::packet::signature::subpacket::NotationDataFlags;
+use openpgp::packet::UserID;
 use openpgp::parse::Parse;
 use openpgp::policy::StandardPolicy;
-use openpgp::serialize::Serialize;
+use openpgp::serialize::{Serialize, SerializeInto};
 
 const P: &StandardPolicy = &StandardPolicy::new();
 
@@ -40,7 +41,8 @@ fn sq_certify() -> Result<()> {
     // A simple certification.
     Command::cargo_bin("sq")
         .unwrap()
-        .arg("certify")
+        .arg("--no-cert-store")
+        .arg("pki").arg("certify")
         .arg(alice_pgp.to_str().unwrap())
         .arg(bob_pgp.to_str().unwrap())
         .arg("bob@example.org")
@@ -48,6 +50,10 @@ fn sq_certify() -> Result<()> {
         .success()
         .stdout(predicate::function(|output: &[u8]| -> bool {
             let cert = Cert::from_bytes(output).unwrap();
+            assert_eq!(cert.bad_signatures().count(), 0,
+                       "Bad signatures in cert\n\n{}",
+                       String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
+
             let vc = cert.with_policy(P, None).unwrap();
 
             for ua in vc.userids() {
@@ -74,15 +80,20 @@ fn sq_certify() -> Result<()> {
     // No expiry.
     Command::cargo_bin("sq")
         .unwrap()
-        .arg("certify")
+        .arg("--no-cert-store")
+        .arg("pki").arg("certify")
         .arg(alice_pgp.to_str().unwrap())
         .arg(bob_pgp.to_str().unwrap())
         .arg("bob@example.org")
-        .args(["--expires", "never"])
+        .args(["--expiry", "never"])
         .assert()
         .success()
         .stdout(predicate::function(|output: &[u8]| -> bool {
             let cert = Cert::from_bytes(output).unwrap();
+            assert_eq!(cert.bad_signatures().count(), 0,
+                       "Bad signatures in cert\n\n{}",
+                       String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
+
             let vc = cert.with_policy(P, None).unwrap();
 
             for ua in vc.userids() {
@@ -108,7 +119,8 @@ fn sq_certify() -> Result<()> {
     // Have alice certify bob@example.org for 0xB0B.
     Command::cargo_bin("sq")
         .unwrap()
-        .arg("certify")
+        .arg("--no-cert-store")
+        .arg("pki").arg("certify")
         .arg(alice_pgp.to_str().unwrap())
         .arg(bob_pgp.to_str().unwrap())
         .arg("bob@example.org")
@@ -118,11 +130,15 @@ fn sq_certify() -> Result<()> {
         .args(["--regex", "b"])
         .arg("--local")
         .arg("--non-revocable")
-        .args(["--expires-in", "1d"])
+        .args(["--expiry", "1d"])
         .assert()
         .success()
         .stdout(predicate::function(|output: &[u8]| -> bool {
             let cert = Cert::from_bytes(output).unwrap();
+            assert_eq!(cert.bad_signatures().count(), 0,
+                       "Bad signatures in cert\n\n{}",
+                       String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
+
             let vc = cert.with_policy(P, None).unwrap();
 
             for ua in vc.userids() {
@@ -150,7 +166,8 @@ fn sq_certify() -> Result<()> {
     // It should fail if the User ID doesn't exist.
     Command::cargo_bin("sq")
         .unwrap()
-        .arg("certify")
+        .arg("--no-cert-store")
+        .arg("pki").arg("certify")
         .arg(alice_pgp.to_str().unwrap())
         .arg(bob_pgp.to_str().unwrap())
         .arg("bob")
@@ -160,7 +177,8 @@ fn sq_certify() -> Result<()> {
     // With a notation.
     Command::cargo_bin("sq")
         .unwrap()
-        .arg("certify")
+        .arg("--no-cert-store")
+        .arg("pki").arg("certify")
         .args(["--notation", "foo", "bar"])
         .args(["--notation", "!foo", "xyzzy"])
         .args(["--notation", "hello@example.org", "1234567890"])
@@ -171,6 +189,9 @@ fn sq_certify() -> Result<()> {
         .success()
         .stdout(predicate::function(|output: &[u8]| -> bool {
             let cert = Cert::from_bytes(output).unwrap();
+            assert_eq!(cert.bad_signatures().count(), 0,
+                       "Bad signatures in cert\n\n{}",
+                       String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
 
             // The standard policy will reject the
             // certification, because it has an unknown
@@ -281,7 +302,8 @@ fn sq_certify_creation_time() -> Result<()>
 
     // Build up the command line.
     let mut cmd = Command::cargo_bin("sq")?;
-    cmd.args(["certify",
+    cmd.args(["--no-cert-store",
+              "pki", "certify",
               &alice_pgp.to_string_lossy(),
               &bob_pgp.to_string_lossy(), bob,
               "--time", iso8601 ]);
@@ -290,6 +312,9 @@ fn sq_certify_creation_time() -> Result<()>
     let stdout = String::from_utf8_lossy(&assertion.get_output().stdout);
 
     let cert = Cert::from_bytes(&*stdout)?;
+    assert_eq!(cert.bad_signatures().count(), 0,
+               "Bad signatures in cert\n\n{}",
+               String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
 
     let vc = cert.with_policy(P, t)?;
 
@@ -363,7 +388,8 @@ fn sq_certify_with_expired_key() -> Result<()>
 
     // Make sure using an expired key fails by default.
     let mut cmd = Command::cargo_bin("sq")?;
-    cmd.args(["certify",
+    cmd.args(["--no-cert-store",
+              "pki", "certify",
               &alice_pgp.to_string_lossy(),
               &bob_pgp.to_string_lossy(), bob ]);
     cmd.assert().failure();
@@ -371,7 +397,8 @@ fn sq_certify_with_expired_key() -> Result<()>
 
     // Try again.
     let mut cmd = Command::cargo_bin("sq")?;
-    cmd.args(["certify",
+    cmd.args(["--no-cert-store",
+              "pki", "certify",
               "--allow-not-alive-certifier",
               &alice_pgp.to_string_lossy(),
               &bob_pgp.to_string_lossy(), bob ]);
@@ -380,6 +407,9 @@ fn sq_certify_with_expired_key() -> Result<()>
     let stdout = String::from_utf8_lossy(&assertion.get_output().stdout);
 
     let cert = Cert::from_bytes(&*stdout)?;
+    assert_eq!(cert.bad_signatures().count(), 0,
+               "Bad signatures in cert\n\n{}",
+               String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
 
     let vc = cert.with_policy(P, None)?;
 
@@ -452,7 +482,8 @@ fn sq_certify_with_revoked_key() -> Result<()>
 
     // Make sure using an expired key fails by default.
     let mut cmd = Command::cargo_bin("sq")?;
-    cmd.args(["certify",
+    cmd.args(["--no-cert-store",
+              "pki", "certify",
               &alice_pgp.to_string_lossy(),
               &bob_pgp.to_string_lossy(), bob ]);
     cmd.assert().failure();
@@ -460,7 +491,8 @@ fn sq_certify_with_revoked_key() -> Result<()>
 
     // Try again.
     let mut cmd = Command::cargo_bin("sq")?;
-    cmd.args(["certify",
+    cmd.args(["--no-cert-store",
+              "pki", "certify",
               "--allow-revoked-certifier",
               &alice_pgp.to_string_lossy(),
               &bob_pgp.to_string_lossy(), bob ]);
@@ -469,6 +501,9 @@ fn sq_certify_with_revoked_key() -> Result<()>
     let stdout = String::from_utf8_lossy(&assertion.get_output().stdout);
 
     let cert = Cert::from_bytes(&*stdout)?;
+    assert_eq!(cert.bad_signatures().count(), 0,
+               "Bad signatures in cert\n\n{}",
+               String::from_utf8(cert.armored().to_vec().unwrap()).unwrap());
 
     let vc = cert.with_policy(P, None)?;
 
@@ -497,3 +532,87 @@ fn sq_certify_with_revoked_key() -> Result<()>
 
     Ok(())
 }
+
+// Certify a certificate in the cert store.
+#[test]
+fn sq_certify_using_cert_store() -> Result<()>
+{
+    let dir = TempDir::new()?;
+
+    let certd = dir.path().join("cert.d").display().to_string();
+    std::fs::create_dir(&certd).expect("mkdir works");
+
+    let alice_pgp = dir.path().join("alice.pgp").display().to_string();
+    let bob_pgp = dir.path().join("bob.pgp").display().to_string();
+
+    // Generate keys.
+    let mut cmd = Command::cargo_bin("sq")?;
+    cmd.args(["--cert-store", &certd,
+              "key", "generate",
+              "--expiry", "never",
+              "--userid", "<alice@example.org>",
+              "--output", &alice_pgp]);
+    cmd.assert().success();
+
+    let alice = Cert::from_file(&alice_pgp)?;
+
+    let mut cmd = Command::cargo_bin("sq")?;
+    cmd.args(["--cert-store", &certd,
+              "key", "generate",
+              "--expiry", "never",
+              "--userid", "<bob@example.org>",
+              "--output", &bob_pgp]);
+    cmd.assert().success();
+
+    let bob = Cert::from_file(&bob_pgp)?;
+
+    // Import bob's (but not alice's!).
+    let mut cmd = Command::cargo_bin("sq")?;
+    cmd.args(["--cert-store", &certd,
+              "cert", "import", &bob_pgp]);
+    cmd.assert().success();
+
+
+    // Have alice certify bob.
+    let mut cmd = Command::cargo_bin("sq")?;
+    cmd.args(["--cert-store", &certd,
+              "pki", "certify",
+              &alice_pgp,
+              &bob.fingerprint().to_string(),
+              "<bob@example.org>"]);
+
+    let output = cmd.output().expect("success");
+    let stdout = String::from_utf8_lossy(&output.stdout);
+    let stderr = String::from_utf8_lossy(&output.stderr);
+    assert!(output.status.success());
+
+    // Make sure the certificate on stdout is bob and that alice
+    // signed it.
+    let parser = CertParser::from_bytes(stdout.as_bytes())
+        .expect("valid");
+    let found = parser.collect::<Result<Vec<Cert>>>()
+        .expect("valid");
+
+    assert_eq!(found.len(), 1,
+               "stdout:\n{}\nstderr:\n{}",
+               stdout, stderr);
+    let found = found.into_iter().next().expect("have one");
+
+    assert_eq!(found.fingerprint(), bob.fingerprint());
+    assert_eq!(found.userids().count(), 1);
+
+    let ua = found.userids().next().expect("have one");
+    let certifications: Vec<_> = ua.certifications().collect();
+    assert_eq!(certifications.len(), 1);
+    let certification = certifications.into_iter().next().unwrap();
+
+    assert_eq!(certification.get_issuers().into_iter().next(),
+               Some(KeyHandle::from(alice.fingerprint())));
+    certification.clone().verify_userid_binding(
+        alice.primary_key().key(),
+        bob.primary_key().key(),
+        &UserID::from("<bob@example.org>"))
+        .expect("valid certification");
+
+    Ok(())
+}
