@@ -26,10 +26,11 @@ mod integration {
 
         // Build up the command line.
         let mut cmd = Command::cargo_bin("sq")?;
-        cmd.args(["key", "generate",
-                  "--creation-time", iso8601,
-                  "--expires", "never",
-                  "--export", &*key_pgp.to_string_lossy()]);
+        cmd.args(["--no-cert-store",
+                  "key", "generate",
+                  "--time", iso8601,
+                  "--expiry", "never",
+                  "--output", &*key_pgp.to_string_lossy()]);
 
         cmd.assert().success();
 
